"""
This module will contain functions for plotting, and for formatting plots, with the goal of
more easily obtaining a consistent style across plots made by different co-authors.
"""

from .plot import showmo, set_style, label_letter, Char

__all__ = ["showmo", "set_style", "label_letter", "Char"]
